@@ -7,7 +7,6 @@
 
 #include "Snapper.hpp"
 #include "Exceptions.hpp"
-#include "Log.hpp"
 #include "Util.hpp"
 #include <regex>
 
@@ -142,14 +141,22 @@ bool Snapper::isReadOnly() {
 }
 
 void Snapper::setDefault() {
-    Util::exec("btrfs subvolume set-default " + std::string(getRoot()));
+    try {
+        callSnapper("modify --default " + snapshotId);
+    } catch (const ExecutionException &e) {
+        Util::exec("btrfs subvolume set-default " + std::string(getRoot()));
+    }
 }
 
 void Snapper::setReadOnly(bool readonly) {
-    std::string boolstr = "true";
-    if (readonly == false)
-        boolstr = "false";
-    Util::exec("btrfs property set " + std::string(getRoot()) + " ro " + boolstr);
+    try {
+        if (readonly == true)
+            callSnapper("modify --read-only " + snapshotId);
+        else
+            callSnapper("modify --read-write " + snapshotId);
+    } catch (const ExecutionException &e) {
+        Util::exec("btrfs property set " + std::string(getRoot()) + " ro " + (readonly ? "true" : "false"));
+    }
 }
 
 /* Helper methods */
