@@ -25,6 +25,8 @@
 #include <limits.h>
 #include <poll.h>
 #include <sched.h>
+#include <selinux/restorecon.h>
+#include <selinux/selinux.h>
 #include <signal.h>
 #include <sys/inotify.h>
 #include <sys/mount.h>
@@ -106,18 +108,13 @@ void Transaction::impl::snapMount() {
     // mount the snapshot directory on a temporary mount point
     char bindTemplate[] = "/tmp/transactional-update-XXXXXX";
     bindDir = mkdtemp(bindTemplate);
-    std::unique_ptr<BindMount> mntBind{new BindMount{bindDir, MS_PRIVATE}};
+    std::unique_ptr<BindMount> mntBind{new BindMount{bindDir, MS_PRIVATE, true}};
     mntBind->setSource(snapshot->getRoot());
     mntBind->mount();
 
     dirsToMount.push_back(std::make_unique<PropagatedBindMount>("/dev"));
     dirsToMount.push_back(std::make_unique<BindMount>("/var/log"));
 
-    std::vector<std::string> customDirs = config.getArray("BINDDIRS");
-    for (auto it = customDirs.begin(); it != customDirs.end(); ++it) {
-        dirsToMount.push_back(std::make_unique<BindMount>(*it));
-    }
-
     Mount mntVar{"/var"};
     if (mntVar.isMount()) {
         if (fs::is_directory("/var/lib/zypp"))
@@ -127,7 +124,44 @@ void Transaction::impl::snapMount() {
             dirsToMount.push_back(std::make_unique<BindMount>("/var/lib/alternatives"));
         if (fs::is_directory("/var/lib/selinux"))
             dirsToMount.push_back(std::make_unique<BindMount>("/var/lib/selinux"));
+        if (is_selinux_enabled()) {
+            // If packages installed files into /var (which is not allowed, but still happens), they will end
+            // up in the root file system, but will always be shadowed by the real /var mount. Due to that they
+            // also won't be relabelled at any time. During updates this may cause problems if packages try to
+            // access those leftover directories with wrong permissions, so they have to be relabelled manually...
+            BindMount selinuxVar("/var/lib/selinux", 0, true);
+            selinuxVar.mount(bindDir);
+            BindMount selinuxEtc("/etc/selinux", 0, true);
+            selinuxEtc.mount(bindDir);
+
+            // restorecon keeps open file handles, so execute it in a child process - umount will fail otherwise
+            pid_t childPid = fork();
+            if (childPid < 0) {
+                throw std::runtime_error{"Forking for SELinux relabelling failed: " + std::string(strerror(errno))};
+            } else if (childPid == 0) {
+                if (chroot(bindDir.c_str()) < 0) {
+                    tulog.error("Chrooting to " + bindDir + " for SELinux relabelling failed: " + std::string(strerror(errno)));
+                    _exit(errno);
+                }
+                unsigned int restoreconOptions = SELINUX_RESTORECON_RECURSE | SELINUX_RESTORECON_IGNORE_DIGEST;
+                if (tulog.level >= TULogLevel::Info)
+                    restoreconOptions |= SELINUX_RESTORECON_VERBOSE;
+                if (selinux_restorecon("/var", restoreconOptions) < 0) {
+                    tulog.error("Relabelling of snapshot /var failed: " + std::string(strerror(errno)));
+                    _exit(errno);
+                }
+                _exit(0);
+            }
+            else {
+                int status;
+                waitpid(childPid, &status, 0);
+                if ((WIFEXITED(status) && WEXITSTATUS(status) != 0) || WIFSIGNALED(status)) {
+                    throw std::runtime_error{"SELinux relabelling failed."};
+                }
+            }
+        }
     }
+
     std::unique_ptr<Mount> mntEtc{new Mount{"/etc"}};
     if (mntEtc->isMount() && mntEtc->getFilesystem() == "overlay") {
         Overlay overlay = Overlay{snapshot->getUid()};
@@ -151,10 +185,12 @@ void Transaction::impl::snapMount() {
     dirsToMount.push_back(std::move(mntVarTmp));
 
     // Mount platform specific GRUB directories for GRUB updates
-    for (auto& path: fs::directory_iterator("/boot/grub2")) {
-        if (fs::is_directory(path)) {
-            if (BindMount{path.path()}.isMount())
-                dirsToMount.push_back(std::make_unique<BindMount>(path.path()));
+    if (fs::exists("/boot/grub2")) {
+        for (auto& path: fs::directory_iterator("/boot/grub2")) {
+            if (fs::is_directory(path)) {
+                if (BindMount{path.path()}.isMount())
+                    dirsToMount.push_back(std::make_unique<BindMount>(path.path()));
+            }
         }
     }
     if (BindMount{"/boot/efi"}.isMount())
@@ -171,6 +207,14 @@ void Transaction::impl::snapMount() {
     if (BindMount{"/boot/writable"}.isMount())
         dirsToMount.push_back(std::make_unique<BindMount>("/boot/writable"));
 
+    std::vector<std::string> customDirs = config.getArray("BINDDIRS");
+    for (auto it = customDirs.begin(); it != customDirs.end(); ++it) {
+        if (fs::is_directory(*it))
+            dirsToMount.push_back(std::make_unique<BindMount>(*it));
+        else
+            tulog.info("Not bind mounting directory '" + *it + "' as it doesn't exist.");
+    }
+
     dirsToMount.push_back(std::make_unique<BindMount>("/.snapshots"));
 
     for (auto it = dirsToMount.begin(); it != dirsToMount.end(); ++it) {
@@ -276,7 +320,7 @@ int Transaction::impl::inotifyRead() {
     struct pollfd pfd = {inotifyFd, POLLIN, 0};
     ret = (poll(&pfd, 1, 500));
     if (ret == -1) {
-        throw std::runtime_error{"Polling inotify file descriptior failed: " + std::string(strerror(errno))};
+        throw std::runtime_error{"Polling inotify file descriptor failed: " + std::string(strerror(errno))};
     } else if (ret > 0) {
         numRead = read(inotifyFd, buf, bufLen);
         if (numRead == 0)
@@ -432,12 +476,33 @@ void Transaction::finalize() {
             (inotifyFd == 0 && fs::exists(getRoot() / "discardIfNoChange")))) {
         tulog.info("No changes to the root file system - discarding snapshot.");
 
-        // Even if the snapshot itself did not contain any changes, /etc may do so. Changes
-        // in /etc may be applied immediately, so merge them back into the running system.
+        // Even if the snapshot itself does not contain any changes, /etc may do so. If the new snapshot is a
+        // direct descendant of the currently running system, then merge the changes back into the currently
+        // running system directly and delete the snapshot. Otherwise merge it back into the previous overlay
+        // (using rsync instead of a plain copy to preserve xattrs).
         std::unique_ptr<Mount> mntEtc{new Mount{"/etc"}};
         if (mntEtc->isMount() && mntEtc->getFilesystem() == "overlay") {
-            Util::exec("rsync --archive --inplace --xattrs --acls --exclude 'fstab' --delete --quiet '" + this->pImpl->bindDir + "/etc/' /etc");
+            std::filesystem::path targetRoot;
+            std::unique_ptr<Mount> previousEtc{new Mount("/etc", 0, true)};
+            if (pImpl->snapshotMgr->getCurrent() == Overlay{pImpl->snapshot->getUid()}.getPreviousSnapshotOvlId()) {
+                tulog.info("Merging changes in /etc into the running system.");
+                targetRoot = "/";
+            } else {
+                tulog.info("Merging changes in /etc into the previous snapshot.");
+
+                auto previousSnapId = Overlay{pImpl->snapshot->getUid()}.getPreviousSnapshotOvlId();
+                std::unique_ptr<Snapshot> previousSnapshot = pImpl->snapshotMgr->open(previousSnapId);
+                previousEtc->setTabSource(previousSnapshot->getRoot() / "etc" / "fstab");
+
+                Overlay previousOvl{previousSnapId};
+                previousOvl.lowerdirs.back() = previousSnapshot->getRoot();
+                previousOvl.setMountOptionsForMount(previousEtc);
+                targetRoot = previousOvl.upperdir.parent_path() / "sync";
+                previousEtc->mount(targetRoot);
+            }
+            Util::exec("rsync --archive --inplace --xattrs --acls --exclude 'fstab' --delete --quiet '" + this->pImpl->bindDir + "/etc/' " + targetRoot.native() + "/etc");
         }
+
         return;
     }
     if (fs::exists(getRoot() / "discardIfNoChange")) {
